// By JK, www.jk-quantized.com

// Setup =================================================================

#include <Adafruit_GFX.h>    // Core graphics library
#include <Adafruit_TFTLCD.h> // Hardware-specific library
#include <TouchScreen.h>
// #include <avr/pgmspace.h>


// Setup LCD
#define LCD_CS A3
#define LCD_CD A2
#define LCD_WR A1
#define LCD_RD A0
#define LCD_RESET A4  // optional

Adafruit_TFTLCD tft(LCD_CS, LCD_CD, LCD_WR, LCD_RD, LCD_RESET);


// Setup touch screen
#define YP A2  // must be an analog pin, use "An" notation!
#define XM A1  // must be an analog pin, use "An" notation!
#define YM 6   // can be a digital pin
#define XP 7   // can be a digital pin

// For better pressure precision, we need to know the resistance
// between X+ and X- Use any multimeter to read it
// For the one we're using, its 300 ohms across the X plate
TouchScreen ts = TouchScreen(XP, YP, XM, YM, 300);


// Touch screen is bigger than LCD. Ignore values beyond these boundaries:
#define TS_MINX 145 //150
#define TS_MINY 120 //120
#define TS_MAXX 920 //920
#define TS_MAXY 940 //940

#define MINPRESSURE 5
#define MAXPRESSURE 1000


// Program specific ======================================================

// Assign human-readable names to some common 16-bit color values:
#define BLACK   0x0000
#define BLUE    0x001F
#define RED     0xF800
#define GREEN   0x07E0
#define CYAN    0x07FF
#define MAGENTA 0xF81F
#define YELLOW  0xFFE0
#define WHITE   0xFFFF

// Paint
#define PENRADIUS 1

// Keyboard
#define COLOR0 0x667C
#define COLOR1 0xA6DA
#define COLOR2 0xE73C
#define COLOR3 0x7D15
#define COLOR4 0xF5AA
#define TXTCOL 0x0841
#define nBtns 58
#define outlineMode true

// Touch screen
#define tMargin 4       // ex. due to stylus radius and weight shifting...
#define samplePeriod 10 // n loops between sampling touch


// Dynamic globals =======================================================


// Touch screen (required)
TSPoint p;
bool validTouch = false;
TSPoint p0;
bool mouseMoved = false;
int tCount = 0;


// Keyboard (required)
int btnMenu = 0;
int btnX, btnY, btnW, btnH, btnKeyId; // as globals cause updated frequently

const PROGMEM uint16_t buttonsN[nBtns][11] = {
//[ color, x, y, w, h, txtX, txtY, txtSize, keyId0, keyId1, keyId2, txt0, txt1, txt2 ]
  { COLOR0,  28,   325,  22,  26,  35,   331,  14,  48,   48,   48   }, //  "0",     "0",     "0"    
  { COLOR0,  52,   325,  22,  26,  59,   331,  14,  49,   49,   49   }, //  "1",     "1",     "1"    
  { COLOR0,  76,   325,  22,  26,  83,   331,  14,  50,   50,   50   }, //  "2",     "2",     "2"    
  { COLOR0,  100,  325,  22,  26,  107,  331,  14,  51,   51,   51   }, //  "3",     "3",     "3"    
  { COLOR0,  124,  325,  22,  26,  131,  331,  14,  52,   52,   52   }, //  "4",     "4",     "4"    
  { COLOR0,  148,  325,  22,  26,  155,  331,  14,  53,   53,   53   }, //  "5",     "5",     "5"    
  { COLOR0,  172,  325,  22,  26,  179,  331,  14,  54,   54,   54   }, //  "6",     "6",     "6"    
  { COLOR0,  196,  325,  22,  26,  203,  331,  14,  55,   55,   55   }, //  "7",     "7",     "7"    
  { COLOR0,  220,  325,  22,  26,  227,  331,  14,  56,   56,   56   }, //  "8",     "8",     "8"    
  { COLOR0,  244,  325,  22,  26,  251,  331,  14,  57,   57,   57   }, //  "9",     "9",     "9"    
  { COLOR1,  4,    356,  22,  26,  11,   361,  15,  113,  126,  81   }, //  "q",     "~",     "Q"    
  { COLOR1,  28,   356,  22,  26,  35,   361,  15,  119,  33,   87   }, //  "w",     "!",     "W"    
  { COLOR1,  52,   356,  22,  26,  59,   361,  15,  101,  64,   69   }, //  "e",     "@",     "E"    
  { COLOR1,  76,   356,  22,  26,  83,   361,  15,  114,  35,   82   }, //  "r",     "#",     "R"    
  { COLOR1,  100,  356,  22,  26,  107,  361,  15,  116,  36,   84   }, //  "t",     "$",     "T"    
  { COLOR1,  124,  356,  22,  26,  131,  361,  15,  121,  37,   89   }, //  "y",     "%",     "Y"    
  { COLOR1,  148,  356,  22,  26,  155,  361,  15,  117,  63,   85   }, //  "u",     "?",     "U"    
  { COLOR1,  172,  356,  22,  26,  179,  361,  15,  105,  92,   73   }, //  "i",     "\\",    "I"    
  { COLOR1,  196,  356,  22,  26,  203,  361,  15,  111,  96,   79   }, //  "o",     "`",     "O"    
  { COLOR1,  220,  356,  22,  26,  227,  361,  15,  112,  58,   80   }, //  "p",     ":",     "P"    
  { COLOR2,  244,  356,  22,  26,  247,  361,  15,  40,   40,   41   }, //  "(",     "(",     ")"    
  { COLOR2,  268,  356,  22,  26,  271,  361,  15,  123,  123,  125  }, //  "{",     "{",     "}"    
  { COLOR2,  292,  356,  22,  26,  295,  361,  15,  91,   91,   93   }, //  "[",     "[",     "]"    
  { COLOR1,  14,   387,  22,  26,  21,   392,  15,  97,   0,    65   }, //  "a",     "",      "A"    
  { COLOR1,  38,   387,  22,  26,  45,   392,  15,  115,  94,   83   }, //  "s",     "^",     "S"    
  { COLOR1,  62,   387,  22,  26,  69,   392,  15,  100,  43,   68   }, //  "d",     "+",     "D"    
  { COLOR1,  86,   387,  22,  26,  93,   392,  15,  102,  45,   70   }, //  "f",     "-",     "F"    
  { COLOR1,  110,  387,  22,  26,  117,  392,  15,  103,  42,   71   }, //  "g",     "*",     "G"    
  { COLOR1,  134,  387,  22,  26,  141,  392,  15,  104,  38,   72   }, //  "h",     "&",     "H"    
  { COLOR1,  158,  387,  22,  26,  165,  392,  15,  106,  124,  74   }, //  "j",     "|",     "J"    
  { COLOR1,  182,  387,  22,  26,  189,  392,  15,  107,  60,   75   }, //  "k",     "<",     "K"    
  { COLOR1,  206,  387,  22,  26,  213,  392,  15,  108,  62,   76   }, //  "l",     ">",     "L"    
  { COLOR2,  230,  387,  22,  26,  237,  392,  15,  59,   59,   59   }, //  ";",     ";",     ";"    
  { COLOR1,  38,   418,  22,  26,  45,   423,  15,  122,  0,    90   }, //  "z",     "",      "Z"    
  { COLOR1,  62,   418,  22,  26,  69,   423,  15,  120,  0,    88   }, //  "x",     "",      "X"    
  { COLOR1,  86,   418,  22,  26,  93,   423,  15,  99,   0,    67   }, //  "c",     "",      "C"    
  { COLOR1,  110,  418,  22,  26,  117,  423,  15,  118,  0,    86   }, //  "v",     "",      "V"    
  { COLOR1,  134,  418,  22,  26,  141,  423,  15,  98,   39,   66   }, //  "b",     "'",     "B"    
  { COLOR1,  158,  418,  22,  26,  165,  423,  15,  110,  95,   78   }, //  "n",     "_",     "N"    
  { COLOR1,  182,  418,  22,  26,  189,  423,  15,  109,  44,   77   }, //  "m",     ",",     "M"    
  { COLOR2,  206,  418,  22,  26,  213,  423,  15,  46,   46,   46   }, //  ".",     ".",     "."    
  { COLOR3,  232,  418,  26,  26,  239,  423,  15,  130,  130,  130  }, //  "◀",     "◀",     "◀"    
  { COLOR3,  261,  395,  26,  26,  268,  400,  15,  131,  131,  131  }, //  "◭",     "◭",     "◭"    
  { COLOR3,  261,  441,  26,  26,  268,  446,  15,  133,  133,  133  }, //  "◮",     "◮",     "◮"    
  { COLOR3,  290,  418,  26,  26,  297,  423,  15,  132,  132,  132  }, //  "▶",     "▶",     "▶"    
  { COLOR4,  3,    325,  19,  26,  3,    332,  11,  140,  140,  140  }, //  "esc",   "esc",   "esc"  
  { COLOR4,  270,  325,  22,  26,  276,  332,  11,  141,  141,  141  }, //  "◁",     "◁",     "◁"    
  { COLOR4,  294,  325,  22,  26,  300,  332,  11,  142,  142,  142  }, //  "▷",     "▷",     "▷"    
  { COLOR2,  292,  387,  22,  26,  299,  392,  15,  61,   61,   61   }, //  "=",     "=",     "="    
  { COLOR4,  4,    418,  31,  26,  12,   423,  15,  153,  153,  153  }, //  "⏶",     "⏶",     "⏶"    
  { COLOR4,  4,    449,  26,  26,  11,   456,  11,  143,  143,  143  }, //  "#&",    "#&",    "#&"   
  { COLOR4,  38,   449,  29,  26,  46,   456,  11,  139,  139,  139  }, //  "⌦",     "⌦",     "⌦"    
  { COLOR4,  71,   449,  25,  26,  74,   456,  11,  127,  127,  127  }, //  "tab",   "tab",   "tab"  
  { COLOR4,  100,  449,  53,  26,  111,  456,  11,  32,   32,   32   }, //  "space", "space", "space"
  { COLOR4,  157,  449,  38,  26,  164,  456,  11,  128,  128,  128  }, //  "entr",  "entr",  "entr" 
  { COLOR4,  199,  449,  29,  26,  206,  456,  11,  129,  129,  129  }, //  "⌫",     "⌫",     "⌫"    
  { COLOR2,  233,  449,  22,  26,  240,  454,  15,  47,   47,   47   }, //  "/",     "/",     "/"    
  { COLOR2,  292,  449,  22,  26,  299,  454,  15,  34,   34,   34   }, //  '"',     '"',     '"'    
};

const PROGMEM char buttonsS[nBtns][3][6] = {
//[ txt0, txt1, txt2 ]
  { "0",     "0",     "0"     },
  { "1",     "1",     "1"     },
  { "2",     "2",     "2"     },
  { "3",     "3",     "3"     },
  { "4",     "4",     "4"     },
  { "5",     "5",     "5"     },
  { "6",     "6",     "6"     },
  { "7",     "7",     "7"     },
  { "8",     "8",     "8"     },
  { "9",     "9",     "9"     },
  { "q",     "~",     "Q"     },
  { "w",     "!",     "W"     },
  { "e",     "@",     "E"     },
  { "r",     "#",     "R"     },
  { "t",     "$",     "T"     },
  { "y",     "%",     "Y"     },
  { "u",     "?",     "U"     },
  { "i",     "\\",    "I"     },
  { "o",     "`",     "O"     },
  { "p",     ":",     "P"     },
  { "(",     "(",     ")"     },
  { "{",     "{",     "}"     },
  { "[",     "[",     "]"     },
  { "a",     "",      "A"     },
  { "s",     "^",     "S"     },
  { "d",     "+",     "D"     },
  { "f",     "-",     "F"     },
  { "g",     "*",     "G"     },
  { "h",     "&",     "H"     },
  { "j",     "|",     "J"     },
  { "k",     "<",     "K"     },
  { "l",     ">",     "L"     },
  { ";",     ";",     ";"     },
  { "z",     "",      "Z"     },
  { "x",     "",      "X"     },
  { "c",     "",      "C"     },
  { "v",     "",      "V"     },
  { "b",     "'",     "B"     },
  { "n",     "_",     "N"     },
  { "m",     ",",     "M"     },
  { ".",     ".",     "."     },
  { "◀",     "◀",     "◀"     },
  { "◭",     "◭",     "◭"     },
  { "◮",     "◮",     "◮"     },
  { "▶",     "▶",     "▶"     },
  { "esc",   "esc",   "esc"   },
  { "◁",     "◁",     "◁"     },
  { "▷",     "▷",     "▷"     },
  { "=",     "=",     "="     },
  { "⏶",     "⏶",     "⏶"     },
  { "#&",    "az",    "#&"    },
  { "⌦",     "⌦",     "⌦"     },
  { "tab",   "tab",   "tab"   },
  { "space", "space", "space" },
  { "entr",  "entr",  "entr"  },
  { "⌫",     "⌫",     "⌫"     },
  { "/",     "/",     "/"     },
  { '"',     '"',     '"'     },
};


// Output keypress to LCD
char btnTxtChar; // as globals cause updated frequently
int btnTxtLen;   // as globals cause updated frequently
const int msgLength = 1200;
char leMessage[msgLength];
int msgIdx = 0;


// 
bool debugMode = false;


// Functions =============================================================


// Keyboard functions ---

void drawKeyboard() {

  char labelChar;
  int labelLength;

  tft.setTextColor( TXTCOL );
  tft.setTextSize(1);

  // Clear the region ---
  tft.fillRect( 0, 320, 320, 480, BLACK );

  // Draw the keyboard ---
  for ( int i = 0; i < nBtns; i += 1 ) {

    // https://www.arduino.cc/en/Reference/PROGMEM

    // Draw the button rect ---
    if ( !outlineMode ) {
      // tft.fillRect( buttonsN[i][1], buttonsN[i][2], buttonsN[i][3], buttonsN[i][4], buttonsN[i][0] );
      tft.fillRect(
        pgm_read_word_near( buttonsN[i] + 1 ),
        pgm_read_word_near( buttonsN[i] + 2 ),
        pgm_read_word_near( buttonsN[i] + 3 ),
        pgm_read_word_near( buttonsN[i] + 4 ),
        pgm_read_word_near( buttonsN[i] + 0 )
      );
    }
    
    // Set position of text cursor ---
    // tft.setCursor( buttonsN[i][5], buttonsN[i][6] );
    tft.setCursor(
      pgm_read_word_near( buttonsN[i] + 5 ) + 1,
      pgm_read_word_near( buttonsN[i] + 6 ) + 4
    );

    // Set the text color ---
    if ( outlineMode ) {
      // tft.setTextColor( buttonsN[i][0] );
      tft.setTextColor( pgm_read_word_near( buttonsN[i] + 0 ) );
    }
    
    // Draw the label ---
    // tft.println( buttonsS[i][btnMenu] );
    labelLength = strlen_P( buttonsS[i][btnMenu] );
    for ( int j = 0; j < labelLength; j++ ) {
      
      labelChar = pgm_read_byte_near( buttonsS[i][btnMenu] + j );
      tft.print( labelChar );  // print character j of the label
    }        
    tft.println(); // end label printing

  }
}


void getKeyPressed() {

  for ( int i = 0; i < nBtns; i += 1 ) {

    // btnX = buttonsN[i][1]; btnY = buttonsN[i][2]; btnW = buttonsN[i][3]; btnH = buttonsN[i][4];
    btnX = pgm_read_word_near( buttonsN[i] + 1 );
    btnY = pgm_read_word_near( buttonsN[i] + 2 );
    btnW = pgm_read_word_near( buttonsN[i] + 3 );
    btnH = pgm_read_word_near( buttonsN[i] + 4 );
    btnKeyId = pgm_read_word_near( buttonsN[i] + 8 + btnMenu );
    
    if( p.x > btnX && p.x < btnX + btnW &&
        p.y > btnY && p.y < btnY + btnH )
    {

      handleKeyPress( btnKeyId, i );
    }
  }
}


void handleKeyPress( int btnKeyId, int btnIdx ) {

  // Toggle symbols
  if ( btnKeyId == 143 ) {

    if ( btnMenu == 0 || btnMenu == 2 ) {

      btnMenu = 1;
      drawKeyboard();
    }
    else if ( btnMenu == 1 ) {

      btnMenu = 0;
      drawKeyboard();
    }
  }

  // Toggle caps
  else if ( btnKeyId == 153 ) {

    if ( btnMenu == 0 || btnMenu == 1 ) {

      btnMenu = 2;
      drawKeyboard();
    }
    else if ( btnMenu == 2 ) {

      btnMenu = 0;
      drawKeyboard();
    }
  }

  // Temp (debug)
  else if ( btnKeyId == 131 ) {
    
    debugMode = !debugMode;
  }

  // Store keyId to register...
  else {

    // Todo..

    // Meanwhile,
    updateMsg( btnIdx );
  }

}


// Touch Screen functions ---

void getTouchCoordinates() {

  TSPoint pt = ts.getPoint();
  
  pinMode( XM, OUTPUT );
  pinMode( YP, OUTPUT );

  if ( pt.z > MINPRESSURE && pt.z < MAXPRESSURE ) {

    // scale from touch 0..1023 to LCD dimensions
    pt.x = map(pt.x, TS_MINX, TS_MAXX, 0, tft.width() );
    pt.y = map(pt.y, TS_MINY, TS_MAXY, 0, tft.height() );

    /*Serial.print( "(" );  Serial.print( pt.x );
    Serial.print( ", " ); Serial.print( pt.y );
    Serial.println( ")" );*/

    p = pt;

    validTouch = true;

    checkIfMouseMoved();

  }
  else {

    validTouch = false;
  }

}

void checkIfMouseMoved() {

  // Assumption, if p is in precisely the same location as it was before,
  //  the mouse/stylus hasn't moved. The chance that a user will click/touch the same
  //  pixel-exact location is very low.

  // Serial.print( "(" );  Serial.print( p.x );
  // Serial.print( ", " ); Serial.print( p.y );
  // Serial.print( "), " );
  // Serial.print( "(" );  Serial.print( p0.x );
  // Serial.print( ", " ); Serial.print( p0.y );
  // Serial.println( ")" );

  if ( tCount > 0 ) {

    mouseMoved = false;
  }
  else if ( p.x > p0.x - tMargin && p.x < p0.x + tMargin && 
            p.y > p0.y - tMargin && p.y < p0.y + tMargin )  
  {

    mouseMoved = false;
    // Serial.println( "same same" );
  }
  else {

    mouseMoved = true;
    // Serial.println( "moved" );
  }

  p0 = p; // store present position as previous

};


void advanceCount() {

  tCount += 1;

  if( tCount > samplePeriod ) {

    tCount = 0;
  }

}


// Paint functions ---

void paint() {

  if ( p.x - PENRADIUS > 0 &&
       p.x + PENRADIUS < tft.width() &&
       p.y - PENRADIUS > 0 &&
       p.y + PENRADIUS < tft.height() )
  {
    tft.fillCircle(p.x, p.y, PENRADIUS, CYAN);
  }

}


// Output keypress to LCD Functions ---

void updateMsg( int btnIdx ) {

  // Add the button's label chars to leMessage
  btnTxtLen = strlen_P( buttonsS[btnIdx][btnMenu] );

  for ( int j = 0; j < btnTxtLen; j++ ) {
    
    btnTxtChar = pgm_read_byte_near( buttonsS[btnIdx][btnMenu] + j );
    
    Serial.print( btnTxtChar );

    leMessage[ msgIdx ] = btnTxtChar;
    msgIdx += 1;

  }
  Serial.println();


  // Reset-ish
  if ( msgIdx >= msgLength ) {

    msgIdx = 0;
  }

}


void displayMsg() {

  // tft.fillRect( 50, 40, 320 - 50, 50, BLUE );

  tft.setTextColor( WHITE );

  tft.setCursor( 20, 50 );

  tft.println( leMessage );

}


// Setup ---

void setup() {

  Serial.begin( 9600 );
  Serial.println( "Keyboard test!" );

  tft.reset();
  tft.begin( tft.readID() );

  tft.fillScreen( BLACK );

  drawKeyboard();

}


// Loop ---

void loop() {

  advanceCount();
  
  getTouchCoordinates();

  if ( validTouch ) {

    // paint();
    if ( debugMode ) { paint(); }

    if ( mouseMoved ) {
      
      getKeyPressed();

      displayMsg();
    }

  }
  
}


